from typing import Any

from fastapi import FastAPI
from opentelemetry import metrics, trace
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import (
    OTLPMetricExporter,
)
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter,
)
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from opentelemetry.instrumentation.httpx import HTTPXClientInstrumentor
from opentelemetry.instrumentation.logging import LoggingInstrumentor
from opentelemetry.metrics import NoOpMeterProvider
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace import NoOpTracerProvider
from typing_extensions import Dict

from .settings import get_settings

tracer: trace.Tracer | None = None
meter: metrics.Meter | None = None


def get_tracer() -> trace.Tracer:
    if tracer is None:
        raise Exception("Tracer is not initialized")
    return tracer


def get_meter() -> metrics.Meter:
    if meter is None:
        raise Exception("Meter is not initialized")
    return meter


def get_resource_attributes() -> Dict[str, Any]:
    resources = Resource.create()
    resources_dict: Dict[str, Any] = {}
    for key in resources.attributes:
        resources_dict[key] = resources.attributes[key]
    settings = get_settings()
    if settings is None:
        raise Exception("Settings are not initialized")
    resources_dict["workspace"] = settings.workspace
    resources_dict["service.name"] = settings.name
    return resources_dict


def get_metrics_exporter() -> OTLPMetricExporter | None:
    settings = get_settings()
    if settings is None:
        raise Exception("Settings are not initialized")
    if not settings.enable_opentelemetry:
        # Return None or a NoOpExporter equivalent
        return None
    return OTLPMetricExporter()


def get_span_exporter() -> OTLPSpanExporter | None:
    settings = get_settings()
    if not settings.enable_opentelemetry:
        return None
    return OTLPSpanExporter()


def instrument_app(app: FastAPI):
    global tracer
    global meter
    settings = get_settings()
    if settings is None:
        raise Exception("Settings are not initialized")

    if not settings.enable_opentelemetry:
        # Use NoOp implementations to stub tracing and metrics
        trace.set_tracer_provider(NoOpTracerProvider())
        tracer = trace.get_tracer(__name__)

        metrics.set_meter_provider(NoOpMeterProvider())
        meter = metrics.get_meter(__name__)
        return

    resource = Resource.create(
        {
            "service.name": settings.name,
            "service.namespace": settings.workspace,
            "service.workspace": settings.workspace,
        }
    )

    # Set up the TracerProvider if not already set
    if not isinstance(trace.get_tracer_provider(), TracerProvider):
        trace_provider = TracerProvider(resource=resource)
        span_processor = BatchSpanProcessor(get_span_exporter())
        trace_provider.add_span_processor(span_processor)
        trace.set_tracer_provider(trace_provider)
        tracer = trace_provider.get_tracer(__name__)
    else:
        tracer = trace.get_tracer(__name__)

    # Set up the MeterProvider if not already set
    if not isinstance(metrics.get_meter_provider(), MeterProvider):
        metrics_exporter = PeriodicExportingMetricReader(get_metrics_exporter())
        meter_provider = MeterProvider(
            resource=resource, metric_readers=[metrics_exporter]
        )
        metrics.set_meter_provider(meter_provider)
        meter = meter_provider.get_meter(__name__)
    else:
        meter = metrics.get_meter(__name__)

    # Only instrument the app when OpenTelemetry is enabled
    FastAPIInstrumentor.instrument_app(
        app=app, tracer_provider=trace.get_tracer_provider(), meter_provider=metrics.get_meter_provider()
    )
    HTTPXClientInstrumentor().instrument(meter_provider=metrics.get_meter_provider())
    LoggingInstrumentor(tracer_provider=trace.get_tracer_provider()).instrument(
        set_logging_format=True
    )



"""Decorators for creating function tools with Beamlit and LangChain integration."""
import asyncio
import functools
from collections.abc import Callable
from logging import getLogger

from fastapi import Request

from beamlit.models import Function, FunctionKit

logger = getLogger(__name__)

def kit(bl_kit: FunctionKit = None, **kwargs: dict) -> Callable:
    """Create function tools with Beamlit and LangChain integration."""

    def wrapper(func: Callable) -> Callable:
        if bl_kit and not func.__doc__ and bl_kit.description:
            func.__doc__ = bl_kit.description
        return func

    return wrapper


def function(*args, function: Function | dict = None, kit=False, **kwargs: dict) -> Callable:
    """Create function tools with Beamlit and LangChain integration."""
    if function is not None and not isinstance(function, dict):
        raise Exception(
            'function must be a dictionary, example: @function(function={"metadata": {"name": "my_function"}})'
        )
    if isinstance(function, dict):
        function = Function(**function)

    def wrapper(func: Callable) -> Callable:
        if function and not func.__doc__ and function.spec and function.spec.description:
            func.__doc__ = function.spec.description

        @functools.wraps(func)
        async def wrapped(*args, **kwargs):
            if len(args) > 0 and isinstance(args[0], Request):
                body = await args[0].json()
                args = [body.get(param) for param in func.__code__.co_varnames[:func.__code__.co_argcount]]
            return await func(*args, **kwargs) if asyncio.iscoroutinefunction(func) else func(*args, **kwargs)
        return wrapped

    return wrapper



from dataclasses import dataclass
from logging import getLogger
from pathlib import Path
from typing import List

import yaml

from beamlit.common.settings import Settings

logger = getLogger(__name__)


@dataclass
class Credentials:
    apiKey: str = ""
    access_token: str = ""
    refresh_token: str = ""
    expires_in: int = 0
    device_code: str = ""
    client_credentials: str = ""


@dataclass
class WorkspaceConfig:
    name: str
    credentials: Credentials


@dataclass
class ContextConfig:
    workspace: str = ""
    environment: str = ""


@dataclass
class Config:
    workspaces: List[WorkspaceConfig] = None
    context: ContextConfig = None

    def __post_init__(self):
        if self.workspaces is None:
            self.workspaces = []
        if self.context is None:
            self.context = ContextConfig()

    def to_json(self) -> dict:
        return {
            "workspaces": [
                {
                    "name": ws.name,
                    "credentials": {
                        "apiKey": ws.credentials.apiKey,
                        "access_token": ws.credentials.access_token,
                        "refresh_token": ws.credentials.refresh_token,
                        "expires_in": ws.credentials.expires_in,
                        "device_code": ws.credentials.device_code,
                        "client_credentials": ws.credentials.client_credentials,
                    },
                }
                for ws in self.workspaces
            ],
            "context": {
                "workspace": self.context.workspace,
                "environment": self.context.environment,
            },
        }


def load_config() -> Config:
    config = Config()
    home_dir = Path.home()
    if home_dir:
        config_path = home_dir / ".beamlit" / "config.yaml"
        if config_path.exists():
            try:
                with open(config_path) as f:
                    data = yaml.safe_load(f)
                    if data:
                        workspaces = []
                        for ws in data.get("workspaces", []):
                            creds = Credentials(**ws.get("credentials", {}))
                            workspaces.append(WorkspaceConfig(name=ws["name"], credentials=creds))
                        config.workspaces = workspaces
                        if "context" in data:
                            config.context = ContextConfig(**data["context"])
            except yaml.YAMLError:
                # Invalid YAML, use empty config
                pass
    return config


def save_config(config: Config):
    home_dir = Path.home()
    if not home_dir:
        raise RuntimeError("Could not determine home directory")

    config_dir = home_dir / ".beamlit"
    config_file = config_dir / "config.yaml"

    config_dir.mkdir(mode=0o700, parents=True, exist_ok=True)
    with open(config_file, "w", encoding="utf-8") as f:
        yaml.dump(config.to_json(), f)


def list_workspaces() -> List[str]:
    config = load_config()
    return [workspace.name for workspace in config.workspaces]


def current_context() -> ContextConfig:
    config = load_config()
    return config.context


def set_current_workspace(workspace_name: str, environment: str):
    config = load_config()
    config.context.workspace = workspace_name
    config.context.environment = environment
    save_config(config)


def load_credentials(workspace_name: str) -> Credentials:
    config = load_config()
    for workspace in config.workspaces:
        if workspace.name == workspace_name:
            return workspace.credentials
    return Credentials()


def load_credentials_from_settings(settings: Settings) -> Credentials:
    return Credentials(
        apiKey=settings.authentication.apiKey,
        client_credentials=settings.authentication.client.credentials,
    )


def create_home_dir_if_missing():
    home_dir = Path.home()
    if not home_dir:
        logger.error("Error getting home directory")
        return

    credentials_dir = home_dir / ".beamlit"
    credentials_file = credentials_dir / "credentials.json"

    if credentials_file.exists():
        logger.warning("You are already logged in. Enter a new API key to overwrite it.")
    else:
        try:
            credentials_dir.mkdir(mode=0o700, parents=True, exist_ok=True)
        except Exception as e:
            logger.error(f"Error creating credentials directory: {e}")


def save_credentials(workspace_name: str, credentials: Credentials):
    create_home_dir_if_missing()
    if not credentials.access_token and not credentials.apiKey:
        logger.info("No credentials to save, error")
        return

    config = load_config()
    found = False

    for i, workspace in enumerate(config.workspaces):
        if workspace.name == workspace_name:
            config.workspaces[i].credentials = credentials
            found = True
            break

    if not found:
        config.workspaces.append(WorkspaceConfig(name=workspace_name, credentials=credentials))

    save_config(config)


def clear_credentials(workspace_name: str):
    config = load_config()
    config.workspaces = [ws for ws in config.workspaces if ws.name != workspace_name]

    if config.context.workspace == workspace_name:
        config.context.workspace = ""

    save_config(config)



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.core_status import CoreStatus
    from ..models.environment_metadata import EnvironmentMetadata
    from ..models.function_spec import FunctionSpec


T = TypeVar("T", bound="Function")


@_attrs_define
class Function:
    """Function

    Attributes:
        metadata (Union[Unset, EnvironmentMetadata]): Environment metadata
        spec (Union[Unset, FunctionSpec]): Function specification
        status (Union[Unset, CoreStatus]): Core status
    """

    metadata: Union[Unset, "EnvironmentMetadata"] = UNSET
    spec: Union[Unset, "FunctionSpec"] = UNSET
    status: Union[Unset, "CoreStatus"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        metadata: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.metadata, Unset):
            metadata = self.metadata.to_dict()

        spec: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.spec, Unset):
            spec = self.spec.to_dict()

        status: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.status, Unset):
            status = self.status.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if metadata is not UNSET:
            field_dict["metadata"] = metadata
        if spec is not UNSET:
            field_dict["spec"] = spec
        if status is not UNSET:
            field_dict["status"] = status

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.core_status import CoreStatus
        from ..models.environment_metadata import EnvironmentMetadata
        from ..models.function_spec import FunctionSpec

        if not src_dict:
            return None
        d = src_dict.copy()
        _metadata = d.pop("metadata", UNSET)
        metadata: Union[Unset, EnvironmentMetadata]
        if isinstance(_metadata, Unset):
            metadata = UNSET
        else:
            metadata = EnvironmentMetadata.from_dict(_metadata)

        _spec = d.pop("spec", UNSET)
        spec: Union[Unset, FunctionSpec]
        if isinstance(_spec, Unset):
            spec = UNSET
        else:
            spec = FunctionSpec.from_dict(_spec)

        _status = d.pop("status", UNSET)
        status: Union[Unset, CoreStatus]
        if isinstance(_status, Unset):
            status = UNSET
        else:
            status = CoreStatus.from_dict(_status)

        function = cls(
            metadata=metadata,
            spec=spec,
            status=status,
        )

        function.additional_properties = d
        return function

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import ast
import json
import os
import sys
import uuid
from logging import getLogger
from typing import Literal

from beamlit.common import slugify
from beamlit.common.settings import Settings, get_settings, init
from beamlit.models import (
    Agent,
    AgentSpec,
    EnvironmentMetadata,
    Flavor,
    Function,
    FunctionSpec,
    Runtime,
)

from .format import arg_to_dict, format_agent_chain, format_parameters
from .parser import Resource, get_description, get_parameters, get_resources

sys.path.insert(0, os.getcwd())
sys.path.insert(0, os.path.join(os.getcwd(), "src"))

random_id = str(uuid.uuid4())[:8]

def get_runtime_image(type: str, name: str) -> str:
    settings = get_settings()
    registry_url = settings.registry_url.replace("https://", "").replace("http://", "")
    image = f"{registry_url}/{settings.workspace}/{type}s/{name}"
    # Generate a random ID to ensure unique image tags
    image = f"{image}:{random_id}"
    return image


def set_default_values(resource: Resource, deployment: Agent | Function):
    settings = get_settings()
    deployment.metadata.workspace = settings.workspace
    deployment.metadata.environment = settings.environment
    if not deployment.metadata.name:
        deployment.metadata.name = resource.name
    if not deployment.metadata.display_name:
        deployment.metadata.display_name = deployment.metadata.name
    if not deployment.spec.description:
        deployment.spec.description = get_description(None, resource)
    if not deployment.spec.runtime:
        deployment.spec.runtime = Runtime()
    if not deployment.spec.runtime.image:
        deployment.spec.runtime.image = get_runtime_image(resource.type, deployment.metadata.name)
    return deployment

def get_beamlit_deployment_from_resource(
    resource: Resource,
) -> Agent | Function:
    """
    Creates a deployment configuration from a resource.

    Args:
        resource (Resource): The resource to create a deployment for

    Returns:
        Agent | Function: The deployment configuration
    """
    for arg in resource.decorator.keywords:
        if arg.arg == "agent":
            if isinstance(arg.value, ast.Dict):
                value = arg_to_dict(arg.value)
                metadata = EnvironmentMetadata(**value.get("metadata", {}))
                spec = AgentSpec(**value.get("spec", {}))
                agent = Agent(metadata=metadata, spec=spec)
                return set_default_values(resource, agent)
        if arg.arg == "function":
            if isinstance(arg.value, ast.Dict):
                value = arg_to_dict(arg.value)
                metadata = EnvironmentMetadata(**value.get("metadata", {}))
                spec = FunctionSpec(**value.get("spec", {}))
                func = Function(metadata=metadata, spec=spec)
                if not func.spec.parameters:
                    func.spec.parameters = get_parameters(resource)
                return set_default_values(resource, func)
    if resource.type == "agent":
        agent = Agent(metadata=EnvironmentMetadata(), spec=AgentSpec())
        return set_default_values(resource, agent)
    if resource.type == "function":
        func = Function(metadata=EnvironmentMetadata(), spec=FunctionSpec())
        func.spec.parameters = get_parameters(resource)
        return set_default_values(resource, func)
    return None


def get_flavors(flavors: list[Flavor]) -> str:
    """
    Converts a list of Flavor objects to JSON string.

    Args:
        flavors (list[Flavor]): List of Flavor objects

    Returns:
        str: JSON string representation of flavors
    """
    if not flavors:
        return "[]"
    return json.dumps([flavor.to_dict() for flavor in flavors])

def get_agent_yaml(
    agent: Agent, functions: list[tuple[Resource, Function]], settings: Settings
) -> str:
    """
    Generates YAML configuration for an agent deployment.

    Args:
        agent (Agent): Agent deployment configuration
        functions (list[tuple[Resource, FunctionDeployment]]): List of associated functions
        settings (Settings): Application settings

    Returns:
        str: YAML configuration string
    """
    template = f"""
apiVersion: beamlit.com/v1alpha1
kind: Agent
metadata:
  name: {slugify(agent.metadata.name)}
  displayName: {agent.metadata.display_name or agent.metadata.name}
  environment: {settings.environment}
  workspace: {settings.workspace}
spec:
  enabled: true
  policies: [{", ".join(agent.spec.policies or [])}]
  functions: [{", ".join([f"{slugify(function.metadata.name)}" for (_, function) in functions])}]
  agentChain: {format_agent_chain(agent.spec.agent_chain)}
  model: {agent.spec.model}
  runtime:
    image: {agent.spec.runtime.image}
"""
    if agent.spec.description:
        template += f"""    description: |
      {agent.spec.description}"""
    return template


def get_function_yaml(function: Function, settings: Settings) -> str:
    """
    Generates YAML configuration for a function deployment.

    Args:
        function (FunctionDeployment): Function deployment configuration
        settings (Settings): Application settings

    Returns:
        str: YAML configuration string
    """
    return f"""
apiVersion: beamlit.com/v1alpha1
kind: Function
metadata:
  name: {slugify(function.metadata.name)}
  displayName: {function.metadata.display_name or function.metadata.name}
  environment: {settings.environment}
spec:
  enabled: true
  policies: [{", ".join(function.spec.policies or [])}]
  description: |
    {function.spec.description}
  parameters: {format_parameters(function.spec.parameters)}
  runtime:
    image: {function.spec.runtime.image}
"""


def dockerfile(
    type: Literal["agent", "function"],
    resource: Resource,
    deployment: Agent | Function,
) -> str:
    """
    Generates Dockerfile content for agent or function deployment.

    Args:
        type (Literal["agent", "function"]): Type of deployment
        resource (Resource): Resource to be deployed
        deployment (Agent | Function): Resource configuration

    Returns:
        str: Dockerfile content
    """
    settings = get_settings()
    if type == "agent":
        module = f"{resource.module.__file__.split('/')[-1].replace('.py', '')}.{resource.module.__name__}"
    else:
        module = f"functions.{resource.module.__file__.split('/')[-1].replace('.py', '')}.{resource.module.__name__}"
    cmd = ["bl", "serve", "--port", "80", "--module", module]
    if type == "agent":
        cmd.append("--remote")
    cmd_str = ",".join([f'"{c}"' for c in cmd])
    return f"""
FROM python:3.12-slim

ARG UV_VERSION="latest"
RUN apt update && apt install -y curl

# Install uv.
COPY --from=ghcr.io/astral-sh/uv:latest /uv /uvx /bin/
RUN curl -fsSL https://raw.githubusercontent.com/beamlit/toolkit/main/install.sh | BINDIR=/bin sh
WORKDIR /beamlit

# Install the application dependencies.
COPY pyproject.toml /beamlit/pyproject.toml
COPY uv.lock /beamlit/uv.lock
RUN uv sync --no-cache

COPY README.m[d] /beamlit/README.md
COPY LICENS[E] /beamlit/LICENSE
COPY {settings.server.directory} /beamlit/src

ENV PATH="/beamlit/.venv/bin:$PATH"

ENTRYPOINT [{cmd_str}]
"""


def generate_beamlit_deployment(directory: str):
    """
    Generates all necessary deployment files for Beamlit agents and functions.

    Args:
        directory (str): Target directory for generated files

    Creates:
        - Agent and function YAML configurations
        - Dockerfiles for each deployment
        - Directory structure for agents and functions
    """
    settings = init()
    logger = getLogger(__name__)
    logger.info(f"Importing server module: {settings.server.module}")
    functions: list[tuple[Resource, Function]] = []
    agents: list[tuple[Resource, Agent]] = []
    for resource in get_resources("agent", settings.server.directory):
        agent = get_beamlit_deployment_from_resource(resource)
        if agent:
            agents.append((resource, agent))
    for resource in get_resources("function", settings.server.directory):
        function = get_beamlit_deployment_from_resource(resource)
        if function:
            functions.append((resource, function))

    agents_dir = os.path.join(directory, "agents")
    functions_dir = os.path.join(directory, "functions")
    # Create directory if it doesn't exist
    os.makedirs(agents_dir, exist_ok=True)
    os.makedirs(functions_dir, exist_ok=True)
    for resource, agent in agents:
        # write deployment file
        agent_dir = os.path.join(agents_dir, agent.metadata.name)
        os.makedirs(agent_dir, exist_ok=True)
        with open(os.path.join(agent_dir, "agent.yaml"), "w") as f:
            content = get_agent_yaml(agent, functions, settings)
            f.write(content)
        # write dockerfile for build
        with open(os.path.join(agent_dir, "Dockerfile"), "w") as f:
            content = dockerfile("agent", resource, agent)
            f.write(content)
        # write destination docker
        with open(os.path.join(agent_dir, "destination.txt"), "w") as f:
            content = agent.spec.runtime.image
            f.write(content)
    for resource, function in functions:
        # write deployment file
        function_dir = os.path.join(functions_dir, function.metadata.name)
        os.makedirs(function_dir, exist_ok=True)
        with open(os.path.join(function_dir, "function.yaml"), "w") as f:
            content = get_function_yaml(function, settings)
            f.write(content)
        # write dockerfile for build
        with open(os.path.join(function_dir, "Dockerfile"), "w") as f:
            content = dockerfile("function", resource, function)
            f.write(content)
        # write destination docker
        with open(os.path.join(function_dir, "destination.txt"), "w") as f:
            content = function.spec.runtime.image
            f.write(content)


class HTTPError(Exception):
    def __init__(self, status_code: int, message: str):
        self.status_code = status_code
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.status_code} {self.message}"

